#include "CvGameCoreDLL.h"
#include "CvMapGenerator.h"
#include "CvMap.h"
#include "CvFractal.h"
#include "CvPlot.h"
#include "CvGlobals.h"
#include "CvRandom.h"
#include "CvGameCoreUtils.h"
#include "CvGameAI.h"

#include "CvGame.h"
#include "CvPlayer.h"
#include "CvPlayerAI.h"
#include "CvTeamAI.h"

#include "NewLearner.h"

NewLearner::NewLearner(void): alpha(0.25), gamma(0.9), eps(0.2), lambda(0.0)
{
}

NewLearner::~NewLearner(void)
{
}

//sets initial parameters and reads in values from external files
void NewLearner::initQ(bool model_, int civNum)
{
	model=model_;

	if(model){
		//TODO: actually add the civNum to the end
		if(civNum==0){
			fileName="msrioM0-81-0.txt";
		}
		else if(civNum==1){
			fileName="msrioM0-81-1.txt";
		}
		else{
			fileName="msrioM0-81.txt";
		}
		
	}
	else{
		if(civNum==0){
			fileName="msrioQ-81-0.txt";
		}
		else if(civNum==1){
			fileName="msrioQ-81-1.txt";
		}
		else{
			fileName="msrioQ-81.txt";
		}
	}
	numStatesVisited=0;
	for(int si=0;si<numStates;si++){
		numActsTaken[si]=0;
	}

	CivilizationTypes civs[MAX_PLAYERS];
	for (int iI = 0; iI < MAX_PLAYERS; iI++)
		{
			if (GET_PLAYER((PlayerTypes)iI).isAlive())
			{
				civs[iI]=GET_PLAYER((PlayerTypes)iI).getCivilizationType();
			}
			//changing the personality
			//0=Barbarian
			//6=Fred
			//7=Gandhi
			//8=Genghis
			//24=Wash
			if(iI==0){

				//choose a player
				//for just Wash and Fred, wash has civ type 0 and fred 1
				if(GET_PLAYER((PlayerTypes)iI).getCivilizationType()==civNum)
				{
					int startAct=QArgMax(0);
					takeAction(startAct);
				}
				else
				{
					exit(1);
				}
			}
	}
	
	//read in the Q-values from the file
	msrioIn.open (fileName);

	if(!msrioIn){
		initFile();
		return;
	}

	std::string line;

	TCHAR szOut[1024];
	sprintf(szOut, "reading numbers\n");
	gDLL->messageControlLog(szOut);
	
	for(int si=0;si<numStates;si++)
	{
		for(int ai=0;ai<numAct;ai++)
		{
			int i=si*numAct+ai;
			float num;
			getline (msrioIn,line);
			sscanf( line.c_str(), "%f", &num );
			QVal[i]=num;
		}
    }
	//read in the model information
	if(model)
	{
		for(int si=0;si<numStates;si++)
		{
			for(int ai=0;ai<numAct;ai++)
			{
				float num;
				getline (msrioIn,line);
				sscanf( line.c_str(), "%f", &num );
				modelRew[si][ai]=num;
			}
		}
		for(int si=0;si<numStates;si++)
		{
			int num;
			getline (msrioIn,line);
			sscanf( line.c_str(), "%d", &num );
			if(num==1){
				statesVisited[si]=true;
				numStatesVisited++;
			}
			else{
				statesVisited[si]=false;
			}
		}

		for(int si=0;si<numStates;si++)
		{
			for(int ai=0;ai<numAct;ai++)
			{
				int num;
				getline (msrioIn,line);
				sscanf( line.c_str(), "%d", &num );
				if(num==1){
					actionsTaken[si][ai]=true;
					numActsTaken[si]++;
				}
				else{
					actionsTaken[si][ai]=false;
				}
			}
		}
		for(int si=0;si<numStates;si++)
		{
			for(int ai=0;ai<numAct;ai++)
			{
				float num;
				getline (msrioIn,line);
				sscanf( line.c_str(), "%f", &num );
				if(num>=0){
					stateFreq[si][ai]=num;
				}
			}
		} 
		for(int si=0;si<numStates;si++)
		{
			for(int ai=0;ai<numAct;ai++)
			{
				for(int si_=0;si_<numStates;si_++)
				{
					float num;
					getline (msrioIn,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq[si][ai][si_]=num;
						if(stateFreq[si][ai]>0){
							transProbs[si][ai][si_]=transFreq[si][ai][si_]/stateFreq[si][ai];
						}
						else{
							transProbs[si][ai][si_]=0.0;
						}
					}
				}
			}
		}
	}
	int num;
	getline (msrioIn,line);
	sscanf( line.c_str(), "%d", &num );
	if(num>=0){
		numWin=num;
	}
	else{
		numWin=0;
	}
	getline (msrioIn,line);
	sscanf( line.c_str(), "%d", &num );
	if(num>=0){
		numPlayed=num;
	}
	else{
		numPlayed=0;
	}
	msrioIn.close();

	//initialize the state and action
	s=0;
	a=0;
	s_=0;
	a_=0;

	lastDiff=0.0;
}

//if no input file is found, set everything to 0
//everything initialized here (except elig[])
void NewLearner::initFile()
{
		for(int si=0;si<numStates;si++)
		{
			for(int ai=0;ai<numAct;ai++)
			{
				int i=si*numAct+ai;
				QVal[i]=0.0;
			}
		}
		if(model)
		{
			for(int si=0;si<numStates;si++)
			{
				statesVisited[si]=false;
				for(int ai=0;ai<numAct;ai++)
				{
					modelRew[si][ai]=0.0;
					actionsTaken[si][ai]=false;
				}
			}
			
			for(int si=0;si<numStates;si++)
			{
				numActsTaken[si]=0;
				for(int ai=0;ai<numAct;ai++)
				{
					stateFreq[si][ai]=0.0;

					for(int si_=0;si_<numStates;si_++)
					{
						transFreq[si][ai][si_]=0.0;
						transProbs[si][ai][si_]=0.0;
					
					}
				}
			}
		}
		numStatesVisited=0;
		numWin=0;
		numPlayed=0;

		//initialize the state and action
		s=0;
		a=0;
		s_=0;
		a_=0;

		lastDiff=0.0;
}


//to end the game: writes out new values and quits
void NewLearner::endGame()
{
	TCHAR szOut[1024];
	sprintf(szOut, "Exiting normally... \n");
	gDLL->messageControlLog(szOut);

	msrioOut.open(fileName);

	if(msrioOut.is_open())
	{
		for(int s=0;s<numStates;s++)
		{
			for(int a=0;a<numAct;a++)
			{
				int i=s*numAct+a;
				msrioOut << QVal[i] << "\n";

				sprintf(szOut, "Q[s=%d,a=%d]=%f\n", s, a, QVal[i]);
				gDLL->messageControlLog(szOut);
						
			}
		}
		if(model){
			for(int si=0;si<numStates;si++)
			{
				for(int ai=0;ai<numAct;ai++)
				{
					msrioOut << modelRew[si][ai] << "\n";
				}
			}
			for(int si=0;si<numStates;si++)
			{
				if(statesVisited[si]==true)
				{
					msrioOut << 1 << "\n";
				}
				else
				{
					msrioOut << 0 << "\n";
				}
			}
			for(int si=0;si<numStates;si++)
			{
				for(int ai=0;ai<numAct;ai++)
				{

					if(actionsTaken[si][ai]==true)
					{
						msrioOut << 1 << "\n";
					}
					else
					{
						msrioOut << 0 << "\n";
					}
				}
			}
			for(int si=0;si<numStates;si++)
			{
				for(int ai=0;ai<numAct;ai++)
				{
					msrioOut << stateFreq[si][ai] << "\n";
				}
			}
			for(int si=0;si<numStates;si++)
			{
				for(int ai=0;ai<numAct;ai++)
				{
					for(int si_=0;si_<numStates;si_++)
					{
						msrioOut << transFreq[si][ai][si_] << "\n";	
					}
				}
			}
		}
		msrioOut << numWin << "\n";	
		msrioOut << ++numPlayed << "\n";	
		msrioOut.flush();
		msrioOut.close();
	}

	exit(1);

}

// The main learning method. Currently, the states are coded as products of features, but this can be changed
//not all the parameters are used, but they are retained for possible future use
void NewLearner::QLearningStep(int maxPop, int maxLand, int maxTech, int maxWonder)
{
	int popScore=GET_PLAYER((PlayerTypes)0).getPopScore();

	int landScore=GET_PLAYER((PlayerTypes)0).getLandScore();

	int techScore=GET_PLAYER((PlayerTypes)0).getTechScore();

	int wonderScore=GET_PLAYER((PlayerTypes)0).getWondersScore();

	int myPow=GET_PLAYER((PlayerTypes)0).getPower();

	AttitudeTypes att=GET_TEAM(GET_PLAYER((PlayerTypes)1).getTeam()).AI_getAttitude((TeamTypes)0);
	int s5=att;

	EraTypes era=GET_PLAYER((PlayerTypes)0).getCurrentEra();
	int s6=era;
	
	//output these values to log file 
	TCHAR szOut[1024];	
	sprintf(szOut, "Population score=  %d of max %d\n", popScore, maxPop);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Land score=  %d of max %d\n", landScore, maxLand);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Tech score=  %d of max %d\n", techScore, maxTech);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Wonder score=  %d of max %d\n", wonderScore, maxWonder);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Power score=  %d vs.  %d\n", myPow, GET_PLAYER((PlayerTypes)1).getPower());
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "your attitude of me is %d\n", att);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "the current era is %d\n", era);
	gDLL->messageControlLog(szOut);

	double myPopPer=popScore/(double)maxPop;
	double myLandPer=landScore/(double)maxLand;
	double myYechPer=techScore/(double)maxTech;
	double myWonderPer=wonderScore/(double)maxWonder;

	int yourPow=GET_PLAYER((PlayerTypes)1).getPower();

	int yourPopScore=GET_PLAYER((PlayerTypes)1).getPopScore();
	int yourLandScore=GET_PLAYER((PlayerTypes)1).getLandScore();

	double remainingLand=myLandPer-(yourLandScore/(double)maxLand);

	//update the state
	int s1=(int)(popScore-yourPopScore);
	if(s1>10)
	{
		s1=2;
	}
	else if(s1<-10)
	{
		s1=0;
	}
	else{
		s1=1;
	}
	int s2=(int)(landScore-yourLandScore);
	if(s2>10)
	{
		s2=2;
	}
	else if(s1<-10)
	{
		s2=0;
	}
	else{
		s2=1;
	}
	int s3=myPow-yourPow;
	if(s3>10)
	{
		s3=2;			
	}
	else if(s3<-10)
	{
		s3=0;
	}
	else{
		s3=1;
	}
	int s4=0;
	if(remainingLand>0.5){
		s4=2;
	}
	else if(remainingLand>.2){
		s4=1;
	}
	else{
		s4=0;
	}

	//update value
	s_=s4*3*3*3+s3*3*3+s2*3+s1;
//	s_=s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;
//	s_=s6*5*3*3*3*3+s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;

	//three different reward possibilities

	//currently set as the difference in score
	double rew = GET_PLAYER((PlayerTypes)0).calculateScore(false, false)-GET_PLAYER((PlayerTypes)1).calculateScore(false, false);
	
	//difference from last step
	//double rew = GET_PLAYER((PlayerTypes)0).calculateScore(false, false)-GET_PLAYER((PlayerTypes)1).calculateScore(false, false)-lastDiff;
	//lastDiff=GET_PLAYER((PlayerTypes)0).calculateScore(false, false)-GET_PLAYER((PlayerTypes)1).calculateScore(false, false);
	
	//set to 0 and 1 for winning only
	//double rew = 0;

	a_=QArgMax(s_);
	int i=s*numAct+a;
	int i_=s_*numAct+a_;
	
	int newA=QEpsGreedy(s_);
	//for eligibility trace
	double delta=rew+gamma*QVal[i_]-QVal[i];
	elig[i]++;
	
	if(lambda==0){
		QVal[i]+=alpha*(rew+gamma*QVal[i_]-QVal[i]);
	}
	//update vals based on eligibility traces - not very efficient
	else if(lambda>0)
	{
		for(int localS=0;localS<numStates;localS++)
		{
			for(int localA=0;localA<numAct;localA++)
			{
				int localI=localS*numAct+localA;
				QVal[localI]+=alpha*delta*elig[localI];
				if(a_==newA){
					elig[localI]*=gamma*lambda;
				}
				else{
					elig[localI]=0;
				}
			}
		}
	}
	if(lambda>1||lambda<0){
		sprintf(szOut, "lambda must be between 0 and 1, exiting...");
		gDLL->messageControlLog(szOut);
		exit(1);
	}

	//model update
	if(model){
		
		stateFreq[s][a]+=1.0;
		transFreq[s][a][s_]+=1.0;
		
		modelRew[s][a]=(modelRew[s][a]*(stateFreq[s][a]-1)+rew)/stateFreq[s][a];
		for(int si_=0;si_<numStates;si_++){
			transProbs[s][a][s_]=transFreq[s][a][s_]/stateFreq[s][a];
		}

		if(statesVisited[s]==false){
			numStatesVisited++;
			if(actionsTaken[s][a]==false){
				numActsTaken[s]++;
			}
		}
		statesVisited[s]=true;
		actionsTaken[s][a]=true;

		Qmodel(25);
	}

	//choose action and update state
	s=s_;
	a=newA;

	takeAction(a);
}

//switch to a personality based on the action--0=Wash, 1=Fred, 2=Gandhi, 3=Genghis
void NewLearner::takeAction(int actNum)
{
	if(actNum==0)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(24);
	}
	else if(actNum==1)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(6);
	}
	else if(actNum==2)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(7);
	}
	else if(actNum==3)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(8);
	}
	else{
		TCHAR szOut[1024];	
		sprintf(szOut, "Fell through else with action %d, exiting!\n", actNum);
		gDLL->messageControlLog(szOut);
		exit(1);
	}
}

//used to test random actions
void NewLearner::randomize()
{
	int rand1=GC.getGameINLINE().getSorenRandNum(numAct, "epsilon greedy 1");
	takeAction(rand1);

}

//find the action that maximizes the value at state s 
int NewLearner::QArgMax(int si)
{
	double maxVal=QVal[si*numAct];
	int maxA=0;
	for(int ai=1;ai<numAct;ai++)
	{
		int i=si*numAct+ai;
		if(QVal[i]>maxVal)
		{ 
			maxVal=QVal[i];
			maxA=ai;
		}
	}
	return maxA;
}


//with prob eps, choose a random action, otherwise choose argmax
int NewLearner::QEpsGreedy(int si)
{
	int rand1=GC.getGameINLINE().getSorenRandNum(100, "epsilon greedy 1");

	if(rand1<=(eps*100)){
		return GC.getGameINLINE().getSorenRandNum(numAct, "epsilon greedy 2");
	}
	else{
		return QArgMax(si);
	}
}

//build and use a model during learning -- Dyna-Q
void NewLearner::Qmodel(int repeatNum)
{
	for(int r=0;r<repeatNum;r++)
	{
		//choose random state that has already been seen
		int randS=chooseRandomS();
		//choose random action that has already been taken in that state
		int randA=chooseRandomA(randS);

		//update from model
		int nextS=getSfromModel(randS, randA);
		double rew=modelRew[randS][randA];

		//update Q
		int aMax=QArgMax(nextS);
		int i=randS*numAct+randA;
		int i_=nextS*numAct+aMax;
		if(i<0||i_<0){
			TCHAR szOut[1024];	
			sprintf(szOut, "indices less than 0 i=%d, i_=%d", i, i_);
			gDLL->messageControlLog(szOut);
			exit(1);	
		}

		QVal[i]+=alpha*(rew+gamma*QVal[i_]-QVal[i]);

	}
}

//get an already seen s' based on the model 
int NewLearner::getSfromModel(int s, int a)
{
	float rand1=GC.getGameINLINE().getSorenRand().getFloat();
	double sum=0.0;

	for(int si_=0;si_<numStates;si_++)
	{	
		sum+=transProbs[s][a][si_];
		if(sum>rand1)
		{
			return si_;
		}
	}

	TCHAR szOut[1024];
	sprintf(szOut, "returning -1 because float was %f and sum was %f\n", rand1, sum);
	gDLL->messageControlLog(szOut);
	return -1;
}

//get a random already seen s
int NewLearner::chooseRandomS()
{
	int rand1=GC.getGameINLINE().getSorenRandNum(numStatesVisited, "random state");

	int count=0;
	for(int si=0;si<numStates;si++)
	{
		if(statesVisited[si]==true)
		{
			count++;
			if(count>rand1)
			{
				return si;
			}
		}
	}
	return -1;
}

//get a randomly already used action at s
int NewLearner::chooseRandomA(int randS){
	
	int rand1=GC.getGameINLINE().getSorenRandNum(numActsTaken[randS], "random action");
	int count=0;
	for(int ai=0;ai<numAct;ai++)
	{
		if(actionsTaken[randS][ai]==true)
		{
			count++;
			if(count>rand1)
			{
				return ai;
			}
		}
	}

	return -1;
}

//called when the game ends - update the q-values
void NewLearner::QComputeFinal(int myScore, int yourScore, bool win)
{
	if(win){
		numWin++;
	}
	int i=s*numAct+a;

	//using difference in score and transitioning to the beginning
	int i_=0;
	double rew=myScore-yourScore;

	//could also use difference from last step
	//double rew=(myScore-yourScore)-lastDiff;

	QVal[i]+=alpha*(rew+gamma*QVal[i_]-QVal[i]);

	/* using a reward of 1 for winning only and elig traces
	double rew=0;
	if(win){
		rew=1.0;
	}
	else{
		rew=-1.0;
	}
	*/
	double delta=rew+gamma-QVal[i];
	elig[i]++;
	//update vals based on eligibility traces - not very efficient
	if(lambda>0){
		for(int localS=0;localS<numStates;localS++){
			for(int localA=0;localA<numAct;localA++){
				int localI=localS*numAct+localA;
				QVal[localI]+=alpha*delta*elig[localI];
			}
		}
	}

	TCHAR szOut[1024];	
	sprintf(szOut, "final reward is  %f\n", rew);
	gDLL->messageControlLog(szOut);

}

//added this to change the personality of the leader
void CvPlayer::changePerson(int person)
{
	
	if (!isBarbarian() && !isMinorCiv())
	{
		if(person<=0||person>GC.getNumLeaderHeadInfos()){
			TCHAR szOut[1024];
			sprintf(szOut, "Person choice out of range  %d of %d \n", person, GC.getNumLeaderHeadInfos());
			gDLL->messageControlLog(szOut);
			exit(1);
		}
		setPersonalityType((LeaderHeadTypes)person);

		TCHAR szOut[1024];
		sprintf(szOut, "Change personality to  %d out of %d \n", person, GC.getNumLeaderHeadInfos());
		gDLL->messageControlLog(szOut);
	}
}