
#include "CvGameCoreDLL.h"
#include "CvMapGenerator.h"
#include "CvMap.h"
#include "CvFractal.h"
#include "CvPlot.h"
#include "CvGlobals.h"
#include "CvRandom.h"
#include "CvGameCoreUtils.h"
#include "CvGameAI.h"

#include "CvGame.h"
#include "CvPlayer.h"
#include "CvPlayerAI.h"
#include "CvTeamAI.h"

#include "NewLearnerFactored.h"

NewLearnerFactored::NewLearnerFactored(void): alpha(0.25), gamma(0.9), eps(0.2), lambda(0.0)
{
}

NewLearnerFactored::~NewLearnerFactored(void)
{
}

//sets initial parameters and reads in values from external files
//note that without the model, this is the same as Q-learning
void NewLearnerFactored::initQ(bool model_, int civNum)
{
	model=model_;

	if(model){
		//TODO: actually add the civNum to the end
		if(civNum==0){
			fileName="msrioFM-81-0.txt";
		}
		else if(civNum==1){
			fileName="msrioFM-81-1.txt";
		}
		else{
			fileName="msrioFM-81.txt";
		}
		
	}
	else{
		if(civNum==0){
			fileName="msrioF-81-0.txt";
		}
		else if(civNum==1){
			fileName="msrioF-81-1.txt";
		}
		else{
			fileName="msrioF-81.txt";
		}
	}

	numStatesFVisited=0;
	for(int si=0;si<numStatesF;si++){
		numActFsTaken[si]=0;
	}

	CivilizationTypes civs[MAX_PLAYERS];
	for (int iI = 0; iI < MAX_PLAYERS; iI++)
		{
			if (GET_PLAYER((PlayerTypes)iI).isAlive())
			{
				civs[iI]=GET_PLAYER((PlayerTypes)iI).getCivilizationType();
			}
			//changing the personality
			//0=Barbarian
			//6=Fred
			//7=Gandhi
			//8=Genghis
			//24=Wash
			if(iI==0){

				//choose a player
				//for just Wash and Fred, wash has civ type 0 and fred 1
				if(GET_PLAYER((PlayerTypes)iI).getCivilizationType()==civNum)
				{
					int startAct=QArgMax(0);
					takeAction(startAct);
				}
				else
				{
					exit(1);
				}
			}
	}
	
	//read in the q-values from the file
	msrioInF.open ("msrioFM-Fred-81.txt");
	
	if(!msrioInF){
		initFile();
		return;
	}

	std::string line;

	TCHAR szOut[1024];
	sprintf(szOut, "reading numbers\n");
	gDLL->messageControlLog(szOut);
	
	for(int si=0;si<numStatesF;si++)
	{
		for(int ai=0;ai<numActF;ai++)
		{
			int i=si*numActF+ai;
			float num;
			getline (msrioInF,line);
			sscanf( line.c_str(), "%f", &num );
			QVal[i]=num;
		}
    }
	//read in the model information
	if(model)
	{
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				float num;
				getline (msrioInF,line);
				sscanf( line.c_str(), "%f", &num );
				modelRew[si][ai]=num;
			}
		}
		for(int si=0;si<numStatesF;si++)
		{
			int num;
			getline (msrioInF,line);
			sscanf( line.c_str(), "%d", &num );
			if(num==1){
				statesVisited[si]=true;
				numStatesFVisited++;
			}
			else{
				statesVisited[si]=false;
			}
		}
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				int num;
				getline (msrioInF,line);
				sscanf( line.c_str(), "%d", &num );
				if(num==1){
					actionsTaken[si][ai]=true;
					numActFsTaken[si]++;
				}
				else{
					actionsTaken[si][ai]=false;
				}
			}
		}
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				float num;
				getline (msrioInF,line);
				sscanf( line.c_str(), "%f", &num );
				if(num>=0){
					stateFreq[si][ai]=num;
				}
			}
		} 

		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				for(int si_=0;si_<3;si_++)
				{
					float num;
					getline (msrioInF,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq1[si][ai][si_]=num;
					}
				}
			}
		}
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				for(int si_=0;si_<3;si_++)
				{
					float num;
					getline (msrioInF,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq2[si][ai][si_]=num;
					}
				}
			}
		}
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				for(int si_=0;si_<3;si_++)
				{
					float num;
					getline (msrioInF,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq3[si][ai][si_]=num;
					}
				}
			}
		}
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				for(int si_=0;si_<3;si_++)
				{
					float num;
					getline (msrioInF,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq4[si][ai][si_]=num;
					}
				}
			}
		}
		/* these features not being used
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				for(int si_=0;si_<5;si_++)
				{
					float num;
					getline (msrioInF,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq5[si][ai][si_]=num;
					}
				}
			}
		}
		/*
		for(int si=0;si<numStatesF;si++)
		{
			for(int ai=0;ai<numActF;ai++)
			{
				for(int si_=0;si_<7;si_++)
				{
					float num;
					getline (msrioInF,line);
					sscanf( line.c_str(), "%f", &num );
					if(num>=0){
						transFreq6[si][ai][si_]=num;
					}
				}
			}
		}
		*/
	}
	int num;
	getline (msrioInF,line);
	sscanf( line.c_str(), "%d", &num );
	if(num>=0){
		numWin=num;
	}
	else{
		numWin=0;
	}
	getline (msrioInF,line);
	sscanf( line.c_str(), "%d", &num );
	if(num>=0){
		numPlayed=num;
	}
	else{
		numPlayed=0;
	}
	msrioInF.close();

	for(s=0;s<numStatesF;s++){
		for(int a=0;a<numActF;a++){
			for(int fi1_=0;fi1_<3;fi1_++){
				for(int fi2_=0;fi2_<3;fi2_++){
					for(int fi3_=0;fi3_<3;fi3_++){
						for(int fi4_=0;fi4_<3;fi4_++){
							//for(int fi5_=0;fi5_<5;fi5_++){
								//for(int fi6_=0;fi6_<7;fi6_++){
								
									//int si_=fi6_*5*3*3*3*3+fi5_*3*3*3*3+fi4_*3*3*3+fi3_*3*3+fi2_*3+fi1_;
									//int si_=fi5_*3*3*3*3+fi4_*3*3*3+fi3_*3*3+fi2_*3+fi1_;
									int si_=fi4_*3*3*3+fi3_*3*3+fi2_*3+fi1_;
									if(stateFreq[s][a]>0){
										//transProbs[s][a][si_]=(transFreq1[s][a][fi1_]/stateFreq[s][a])*(transFreq2[s][a][fi2_]/stateFreq[s][a])*(transFreq3[s][a][fi3_]/stateFreq[s][a])*(transFreq4[s][a][fi4_]/stateFreq[s][a])*(transFreq5[s][a][fi5_]/stateFreq[s][a])*(transFreq6[s][a][fi6_]/stateFreq[s][a]);
										//transProbs[s][a][si_]=(transFreq1[s][a][fi1_]/stateFreq[s][a])*(transFreq2[s][a][fi2_]/stateFreq[s][a])*(transFreq3[s][a][fi3_]/stateFreq[s][a])*(transFreq4[s][a][fi4_]/stateFreq[s][a])*(transFreq5[s][a][fi5_]/stateFreq[s][a]);
										transProbs[s][a][si_]=(transFreq1[s][a][fi1_]/stateFreq[s][a])*(transFreq2[s][a][fi2_]/stateFreq[s][a])*(transFreq3[s][a][fi3_]/stateFreq[s][a])*(transFreq4[s][a][fi4_]/stateFreq[s][a]);							
									}
									else{
										transProbs[s][a][si_]=0.0;
									}
								//}
							//}
						}
					}
				}
			}
		}
	}
	//initialize the other variables

	s=0;
	s_=0;
	f1=0;
	f2=0;
	f3=0;
	f4=0;
	f5=0;
	f6=0;
	f1_=0;
	f2_=0;
	f3_=0;
	f4_=0;
	f5_=0;
	f6_=0;
	a=0;
	a_=0;
	lastDiff=0;
}

//if no input file is found, set everything to 0
//everything initialized here (except elig[])
void NewLearnerFactored::initFile()
{
		//loops hard coded for now...
		for(int s1=0;s1<3;s1++){
			for(int s2=0;s2<3;s2++){
				for(int s3=0;s3<3;s3++){
					for(int s4=0;s4<3;s4++){
						//for(int s5=0;s5<5;s5++){ 
							//for(int s6=0;s6<7;s6++){
	
							int s=s4*3*3*3+s3*3*3+s2*3+s1;
							//int s=s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;
							//int s=s6*5*3*3*3*3+s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;
								for(int a=0;a<numActF;a++){
									int i=s*numActF+a;
									QVal[i]=0.0;
								}
							//}
						//}
					}
				}
			}
		}
		if(model){
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					modelRew[si][ai]=0.0;
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				statesVisited[si]=false;
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					actionsTaken[si][ai]=false;
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					stateFreq[si][ai]=0.0;
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						transFreq1[si][ai][si_]=0.0;
					
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						transFreq2[si][ai][si_]=0.0;
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						transFreq3[si][ai][si_]=0.0;
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						transFreq4[si][ai][si_]=0.0;
					}
				}
			}
			/* these futures not currently used
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<5;si_++)
					{
						transFreq5[si][ai][si_]=0.0;
					}
				}
			}
			/*
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<7;si_++)
					{
						transFreq6[si][ai][si_]=0.0;
					}
				}
			}
			*/
		}
	for(s=0;s<numStatesF;s++){
		for(int a=0;a<numActF;a++){
			for(int si_=0;si_<numStatesF;si_++){
				transProbs[s][a][si_]=0.0;
			}				
		}
	}
	numWin=0;
	numPlayed=0;

	//initialize other variables
	s=0;
	s_=0;
	f1=0;
	f2=0;
	f3=0;
	f4=0;
	f5=0;
	f6=0;
	f1_=0;
	f2_=0;
	f3_=0;
	f4_=0;
	f5_=0;
	f6_=0;
	a=0;
	a_=0;
	lastDiff=0;

}

//to end the game: writes out new values and quits
void NewLearnerFactored::endGame()
{
	TCHAR szOut[1024];
	sprintf(szOut, "Exiting normally... \n");
	gDLL->messageControlLog(szOut);

	msrioOutF.open (fileName);
	if(msrioOutF.is_open())
	{
		//loops hard coded for now...
		for(int s1=0;s1<3;s1++){
			for(int s2=0;s2<3;s2++){
				for(int s3=0;s3<3;s3++){
					for(int s4=0;s4<3;s4++){
					//	for(int s5=0;s5<5;s5++){ 
						//	for(int s6=0;s6<7;s6++){
	
							int s=s4*3*3*3+s3*3*3+s2*3+s1;
							//int s=s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;
							//int s=s6*5*3*3*3*3+s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;
								for(int a=0;a<numActF;a++){
									int i=s*numActF+a;
									msrioOutF << QVal[i] << "\n";

									sprintf(szOut, "Q[s=%d,%d,%d,%d,a=%d]=%f\n", s1, s2, s3, s4, a, QVal[i]);
									//sprintf(szOut, "Q[s=%d,%d,%d,%d,%d,a=%d]=%f\n", s1, s2, s3, s4, s5, a, QVal[i]);
									//sprintf(szOut, "Q[s=%d,%d,%d,%d,%d,%d,a=%d]=%f\n", s1, s2, s3, s4, s5, s6, a, QVal[i]);
									gDLL->messageControlLog(szOut);
								}
							//}
						//}
					}
				}
			}
		}
		if(model){
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					msrioOutF << modelRew[si][ai] << "\n";
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				if(statesVisited[si]==true)
				{
					msrioOutF << 1 << "\n";
				}
				else
				{
					msrioOutF << 0 << "\n";
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{

					if(actionsTaken[si][ai]==true)
					{
						msrioOutF << 1 << "\n";
					}
					else
					{
						msrioOutF << 0 << "\n";
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					msrioOutF << stateFreq[si][ai] << "\n";
				}
			}
			
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						msrioOutF << transFreq1[si][ai][si_] << "\n";	
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						msrioOutF << transFreq2[si][ai][si_] << "\n";	
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						msrioOutF << transFreq3[si][ai][si_] << "\n";	
					}
				}
			}
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<3;si_++)
					{
						msrioOutF << transFreq4[si][ai][si_] << "\n";	
					}
				}
			}
			/* features not currently being used
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<5;si_++)
					{
						msrioOutF << transFreq5[si][ai][si_] << "\n";	
					}
				}
			}
			/*
			for(int si=0;si<numStatesF;si++)
			{
				for(int ai=0;ai<numActF;ai++)
				{
					for(int si_=0;si_<7;si_++)
					{
						msrioOutF << transFreq6[si][ai][si_] << "\n";	
					}
				}
			}
			*/
		}
		msrioOutF << numWin << "\n";	
		msrioOutF << ++numPlayed << "\n";	
		msrioOutF.flush();
		msrioOutF.close();
	}

	exit(1);
}

// The main learning method.
//not all the parameters are used, but they are retained for possible future use
void NewLearnerFactored::QLearningStep(int maxPop, int maxLand, int maxTech, int maxWonder)
{
	int popScore=GET_PLAYER((PlayerTypes)0).getPopScore();

	int landScore=GET_PLAYER((PlayerTypes)0).getLandScore();
		
	int techScore=GET_PLAYER((PlayerTypes)0).getTechScore();

	int wonderScore=GET_PLAYER((PlayerTypes)0).getWondersScore();

	int myPow=GET_PLAYER((PlayerTypes)0).getPower();	

	AttitudeTypes att=GET_TEAM(GET_PLAYER((PlayerTypes)1).getTeam()).AI_getAttitude((TeamTypes)0);
	int s5=att;

	EraTypes era=GET_PLAYER((PlayerTypes)0).getCurrentEra();
	int s6=era;

	//output these values to log file
	TCHAR szOut[1024];	
	sprintf(szOut, "Population score=  %d of max %d\n", popScore, maxPop);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Land score=  %d of max %d\n", landScore, maxLand);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Tech score=  %d of max %d\n", techScore, maxTech);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Wonder score=  %d of max %d\n", wonderScore, maxWonder);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "Power score=  %d vs.  %d\n", myPow, GET_PLAYER((PlayerTypes)1).getPower());
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "your attitude of me is %d\n", att);
	gDLL->messageControlLog(szOut);

	sprintf(szOut, "the current era is %d\n", era);
	gDLL->messageControlLog(szOut);

	double myPopPer=popScore/(double)maxPop;
	double myLandPer=landScore/(double)maxLand;
	double myYechPer=techScore/(double)maxTech;
	double myWonderPer=wonderScore/(double)maxWonder;

	int yourPow=GET_PLAYER((PlayerTypes)1).getPower();

	int yourPopScore=GET_PLAYER((PlayerTypes)1).getPopScore();
	int yourLandScore=GET_PLAYER((PlayerTypes)1).getLandScore();

	double remainingLand=myLandPer-(yourLandScore/(double)maxLand);

	//update the state
	int s1=(int)(popScore-yourPopScore);
	if(s1>10)
	{
		s1=2;
	}
	else if(s1<-10)
	{
		s1=0;
	}
	else{
		s1=1;
	}
	int s2=(int)(landScore-yourLandScore);
	if(s2>10)
	{
		s2=2;
	}
	else if(s1<-10)
	{
		s2=0;
	}
	else{
		s2=1;
	}
	int s3=myPow-yourPow;
	if(s3>10)
	{
		s3=2;		
	}
	else if(s3<-10)
	{
		s3=0;
	}
	else{
		s3=1;
	}
	int s4=0;
	if(remainingLand>0.5){
		s4=2;
	}
	else if(remainingLand>.2){
		s4=1;
	}
	else{
		s4=0;
	}

	s_=s4*3*3*3+s3*3*3+s2*3+s1;
//	s_=s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;
//	s_=s6*5*3*3*3*3+s5*3*3*3*3+s4*3*3*3+s3*3*3+s2*3+s1;

	f1_=s1;
	f2_=s2;
	f3_=s3;
	f4_=s4;
	f5_=s5;
	f6_=s6;

	//three different reward possibilities

	//currently set as the difference in score
	double rew = GET_PLAYER((PlayerTypes)0).calculateScore(false, false)-GET_PLAYER((PlayerTypes)1).calculateScore(false, false);

	//difference from last step
	//double rew = GET_PLAYER((PlayerTypes)0).calculateScore(false, false)-GET_PLAYER((PlayerTypes)1).calculateScore(false, false)-lastDiff;
	//lastDiff=GET_PLAYER((PlayerTypes)0).calculateScore(false, false)-GET_PLAYER((PlayerTypes)1).calculateScore(false, false);

	//set to 0 and 1 for winning only
	//double rew = 0;

	a_=QArgMax(s_);
	int i=s*numActF+a;
	int i_=s_*numActF+a_;
	
	int newA=QEpsGreedy(s_);
	//for eligibility trace
	double delta=rew+gamma*QVal[i_]-QVal[i];
	elig[i]++;
	
	if(lambda==0){
		QVal[i]+=alpha*(rew+gamma*QVal[i_]-QVal[i]);
	}
	//update vals based on eligibility traces - not very efficient
	else if(lambda>0)
	{
		for(int localS=0;localS<numStatesF;localS++)
		{
			for(int localA=0;localA<numActF;localA++)
			{
				int localI=localS*numActF+localA;
				QVal[localI]+=alpha*delta*elig[localI];
				if(a_==newA){
					elig[localI]*=gamma*lambda;
				}
				else{
					elig[localI]=0;
				}
			}
		}
	}
	if(lambda>1||lambda<0){
		sprintf(szOut, "lambda must be between 0 and 1, exiting...");
		gDLL->messageControlLog(szOut);
		exit(1);
	}

	//model update
	if(model){
		
		stateFreq[s][a]+=1.0;

		transFreq1[s][a][f1_]+=1.0;
		transFreq2[s][a][f2_]+=1.0;
		transFreq3[s][a][f3_]+=1.0;
		transFreq4[s][a][f4_]+=1.0;
		transFreq5[s][a][f5_]+=1.0;
		transFreq6[s][a][f6_]+=1.0;
		
		modelRew[s][a]=(modelRew[s][a]*(stateFreq[s][a]-1)+rew)/stateFreq[s][a];
		for(int fi1_=0;fi1_<3;fi1_++){
			for(int fi2_=0;fi2_<3;fi2_++){
				for(int fi3_=0;fi3_<3;fi3_++){
					for(int fi4_=0;fi4_<3;fi4_++){
				//		for(int fi5_=0;fi5_<5;fi5_++){
					//		for(int fi6_=0;fi6_<7;fi6_++){
							//	int si_=fi6_*5*3*3*3*3+fi5_*3*3*3*3+fi4_*3*3*3+fi3_*3*3+fi2_*3+fi1_;
							//	int si_=fi5_*3*3*3*3+fi4_*3*3*3+fi3_*3*3+fi2_*3+fi1_;
								int si_=fi4_*3*3*3+fi3_*3*3+fi2_*3+fi1_;
							//	transProbs[s][a][si_]=(transFreq1[s][a][fi1_]/stateFreq[s][a])*(transFreq2[s][a][fi2_]/stateFreq[s][a])*(transFreq3[s][a][fi3_]/stateFreq[s][a])*(transFreq4[s][a][fi4_]/stateFreq[s][a])*(transFreq5[s][a][fi5_]/stateFreq[s][a])*(transFreq6[s][a][fi6_]/stateFreq[s][a]);
							//	transProbs[s][a][si_]=(transFreq1[s][a][fi1_]/stateFreq[s][a])*(transFreq2[s][a][fi2_]/stateFreq[s][a])*(transFreq3[s][a][fi3_]/stateFreq[s][a])*(transFreq4[s][a][fi4_]/stateFreq[s][a])*(transFreq5[s][a][fi5_]/stateFreq[s][a]);
								transProbs[s][a][si_]=(transFreq1[s][a][fi1_]/stateFreq[s][a])*(transFreq2[s][a][fi2_]/stateFreq[s][a])*(transFreq3[s][a][fi3_]/stateFreq[s][a])*(transFreq4[s][a][fi4_]/stateFreq[s][a]);				
					//		}
				//		}
					}
				}
			}
		}
		
		if(statesVisited[s]==false){
			numStatesFVisited++;
			if(actionsTaken[s][a]==false){
				numActFsTaken[s]++;
			}
		}

		statesVisited[s]=true;
		actionsTaken[s][a]=true;

		Qmodel(25);
	}

	//choose action and update state
	s=s_;
	f1=f1_;
	f2=f2_;
	f3=f3_;
	f4=f4_;
	f5=f5_;
	f6=f6_;
	a=newA;

	takeAction(a);
}

//switch to a personality based on the action--0=Wash, 1=Fred, 2=Gandhi, 3=Genghis
void NewLearnerFactored::takeAction(int actNum)
{
	if(actNum==0)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(24);
	}
	else if(actNum==1)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(6);
	}
	else if(actNum==2)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(7);
	}
	else if(actNum==3)
	{
		GET_PLAYER((PlayerTypes)0).changePerson(8);
	}
	else{
		TCHAR szOut[1024];	
		sprintf(szOut, "Fell through else with action %d, exiting!\n", actNum);
		gDLL->messageControlLog(szOut);
		exit(1);
	}
}

//used to test random actions
void NewLearnerFactored::randomize()
{
	int rand1=GC.getGameINLINE().getSorenRandNum(numActF, "epsilon greedy 1");
	takeAction(rand1);

}

//find the action that maximizes the value at state s 
int NewLearnerFactored::QArgMax(int s)
{
	double maxVal=QVal[s*numActF];
	int maxA=0;
	for(int ai=1;ai<numActF;ai++)
	{
		int i=s*numActF+ai;
		if(QVal[i]>maxVal)
		{ 
			maxVal=QVal[i];
			maxA=ai;
		}
	}
	return maxA;
}


//with prob eps, choose a random action, otherwise choose argmax
int NewLearnerFactored::QEpsGreedy(int s)
{
	int rand1=GC.getGameINLINE().getSorenRandNum(100, "epsilon greedy 1");

	if(rand1<=(eps*100)){
		return GC.getGameINLINE().getSorenRandNum(numActF, "epsilon greedy 2");
	}
	else{
		return QArgMax(s);
	}
}

//build and use a model during learning -- Dyna-Q
void NewLearnerFactored::Qmodel(int repeatNum)
{
	for(int r=0;r<repeatNum;r++)
	{
		//choose random state that has already been seen
		int randS=chooseRandomS();
		//choose random action that has already been taken in that state
		int randA=chooseRandomA(randS);

		//update from model
		int nextS=getSfromModel(randS, randA);
		double rew=modelRew[randS][randA];

		//update Q
		int aMax=QArgMax(nextS);
		int i=randS*numActF+randA;
		int i_=nextS*numActF+aMax;
		if(i<0||i_<0){
			TCHAR szOut[1024];	
			sprintf(szOut, "indices less than 0 i=%d, i_=%d", i, i_);
			gDLL->messageControlLog(szOut);
			exit(1);	
		}

		QVal[i]+=alpha*(rew+gamma*QVal[i_]-QVal[i]);
	}
}

//get an already seen s' based on the model 
int NewLearnerFactored::getSfromModel(int s, int a)
{
	float rand1=GC.getGameINLINE().getSorenRand().getFloat();
	double sum=0.0;

	for(int si_=0;si_<numStatesF;si_++)
	{
		sum+=transProbs[s][a][si_];
		if(sum>rand1)
		{
			return si_;
		}
	}

	TCHAR szOut[1024];
	sprintf(szOut, "returning -1 because float was %f and sum was %f\n", rand1, sum);
	gDLL->messageControlLog(szOut);
	return -1;
}

//get a random already seen s
int NewLearnerFactored::chooseRandomS()
{
	int rand1=GC.getGameINLINE().getSorenRandNum(numStatesFVisited, "random state");
	int count=0;
	for(int si=0;si<numStatesF;si++)
	{
		if(statesVisited[si]==true)
		{
			count++;
			if(count>rand1)
			{
				return si;
			}
		}
	}
	return -1;
}

//get a randomly already used action at s
int NewLearnerFactored::chooseRandomA(int randS){
	
	int rand1=GC.getGameINLINE().getSorenRandNum(numActFsTaken[randS], "random action");
	int count=0;
	for(int ai=0;ai<numActF;ai++)
	{
		if(actionsTaken[randS][ai]==true)
		{
			count++;
			if(count>rand1)
			{
				return ai;
			}
		}
	}

	return -1;
}

//called when the game ends - update the q-values
void NewLearnerFactored::QComputeFinal(int myScore, int yourScore, bool win)
{

	if(win){
		numWin++;
	}
	int i=s*numActF+a;

	//using difference in score and transitioning to the beginning
	int i_=0;
	double rew=myScore-yourScore;

	//could also use difference from last step
	//double rew=(myScore-yourScore)-lastDiff;

	QVal[i]+=alpha*(rew+gamma*QVal[i_]-QVal[i]);

	/* using a reward of 1 for winning only and elig traces
	double rew=0;
	if(win){
		rew=1.0;
	}
	else{
		rew=-1.0;
	}

	*/
	double delta=rew+gamma-QVal[i];
	elig[i]++;
	//update vals based on eligibility traces - not very efficient
	if(lambda>0){
		for(int localS=0;localS<numStatesF;localS++){
			for(int localA=0;localA<numActF;localA++){
				int localI=localS*numActF+localA;
				QVal[localI]+=alpha*delta*elig[localI];
			}
		}
	}
	TCHAR szOut[1024];	
	sprintf(szOut, "final reward is  %f\n", rew);
	gDLL->messageControlLog(szOut);

}